$ on control_y then goto final_exit
$ on error     then continue
$!
$!***************************
$! Process Arguments
$!***************************
$ tcpip = p1
$ make_type = p2
$ make_options = p3
$ if (make_options .nes. "") then make_options = ",''make_options'"
$!
$! uncomment the next line to remove the uidl support
$ uidl = ",UIDL_EXTEND"
$!
$!***************************
$! Initialize
$!***************************
$ say     := write SYS$OUTPUT
$ facility = "%IUPOP3MAKE"
$!
$ is_alpha = (f$getsyi("HW_MODEL") .gt. 1024)
$ if (is_alpha)
$ then
$   arch = "ALPHA"
$   cc_options  == - 
    "/STANDARD=VAXC/PREFIX=EXCEPT=(socket,bind,listen,accept,getpeername,setsockopt)"
$   create share.opt
!this space intentionally left blank
$ else
$   arch = "VAX"
$   cc_options == ""
$   create share.opt
sys$share:vaxcrtl/share
$ endif
$!
$ vms_version = f$getsyi("VERSION")
$ if ((.not. is_alpha) .and. (vms_version .lts. "V5.4"))
$ then
$   password_version = "53"
$ else
$   password_version = "54"
$ endif
$ say facility,"-I-PASS, making IUPOP3 with password version ",password_version
$!
$!
$!***************************
$! Main Program
$!***************************
$ gosub check_parameters
$ gosub prepare_'tcpip'
$ gosub compile_'make_type'
$!
$ say facility,"-I-SUCCESS, IUPOP3 successfully made"
$ if (tcpip.eqs. "MULTINET")
$ then
$   say ""
$   say "IMPORTANT: You may need to remove Multinet's POP3 server from the"
$   say "           network configuration, before you can start IUPOP3."
$   say ""
$ endif
$ goto  final_exit
$!
$!***************************
$! Check Parameters
$!***************************
$ check_parameters:
$ if (tcpip .eqs. "")
$ then
$   if (f$trnlnm("UCX$DEVICE") .nes. "") then tcpip = "UCX"
$   if (f$trnlnm("MULTINET")   .nes. "") then tcpip = "MULTINET"
$   if (f$trnlnm("TWG$TCP")    .nes. "") then tcpip = "WINS"
$ endif
$!
$ if ((tcpip .eqs. "WINS") .or. -
      (tcpip .eqs. "UCX")  .or. -
      (tcpip .eqs. "MULTINET"))
$   then
$     say facility,"-I-TCPTYPE, making IUPOP3 for ",tcpip, " TCP/IP"
$   else
$     say facility,"-F-INVTCP, '",tcpip,"' TCP/IP is not supported"
$     exit
$ endif
$!
$ if (make_type .eqs. "")
$ then
$   on error then continue
$   define sys$output nl:
$   define sys$error nl:
$   mms
$   ss_status = $status
$   deassign sys$output 
$   deassign sys$error
$   on error then continue
$   if (ss_status .eq. %X00038090)
$   then
$     make_type = "NOMMS"
$   else
$     make_type = "MMS"
$   endif
$ endif
$!
$ if ((make_type .eqs. "MMS") .or. (make_type .eqs. "NOMMS"))
$ then
$   say facility,"-I-MAKETYPE, making IUPOP3 with ",make_type
$ else
$   say facility,"-F-INVMAKE, make type must be MMS or NOMMS"
$   exit
$ endif
$!
$ return
$!
$!***************************
$! Compile MMS
$!***************************
$ compile_mms:
$ mms/description=iupop3.mms
$ return
$!
$!***************************
$! Compile NOMMS
$!***************************
$ compile_nomms:
$ compile := cc/list/opt/define=('tcpip' 'uidl','arch' 'make_options') 'cc_options'
$ verify = f$verify(1)
$ 'compile' iupop3
$ 'compile' iupop3_commands
$ 'compile' iupop3_utility
$ 'compile' iupop3_vms
$ 'compile' passwd_v'password_version'
$ macro/list change_case
$ link -
    iupop3, -
    iupop3_commands, -
    iupop3_utility, -
    iupop3_vms, -
    passwd_v'password_version', -
    change_case, -
    share.opt/opt, -
    'tcpip'.opt/opt
$ verify = f$verify('verify')
$ return
$!
$!***************************
$! Prepare UCX
$!***************************
$ prepare_ucx:
$ define/nolog vaxc$include sys$common:[syslib]
$ define/nolog sys          sys$common:[syslib]
$ define/nolog netinet      sys$common:[syslib]
$ define/nolog arpa         sys$common:[syslib]
$ define/nolog vms          sys$common:[syslib]
$!
$ if (is_alpha)
$ then
$   cc_options == "/STANDARD=VAXC/PREFIX=ALL_ENTRIES"
$   create ucx.opt
!this lines intentionally commented for AXP/UCX; uncomment the next one for VAX
!sys$library:ucx$ipc.olb/lib
$ endif
$!
$ return
$!
$!***************************
$! Prepare WINS
$!***************************
$ prepare_wins:
$ define/nolog vaxc$include sys$common:[syslib], twg$tcp:[netdist.include]
$ define/nolog sys          twg$tcp:[netdist.include.sys]
$ define/nolog netinet      twg$tcp:[netdist.include.netinet]
$ define/nolog arpa         twg$tcp:[netdist.include.arpa]
$ define/nolog vms          twg$tcp:[netdist.include.vms]
$ return
!$
$!***************************
$! Prepare MULTINET
$!***************************
$ prepare_multinet:
$ define/nolog vaxc$include multinet_root:[multinet.include], -
                            sys$common:[syslib]
$ define/nolog sys     multinet_root:[multinet.include.sys], sys$common:[syslib]
$ define/nolog arpa    multinet_root:[multinet.include.arpa]
$ define/nolog netinet multinet_root:[multinet.include.netinet]
$ define/nolog vms     multinet_root:[multinet.include.vms]
$ return
$!
$!***************************
$! Final Exit
$!***************************
$ final_exit:
$ if (f$search("share.opt") .nes "") then delete share.opt;*
$ exit
